from __future__ import annotations

from abc import abstractmethod, ABC
from typing import Tuple, Any, Union

import nip.parser as parser


class Token(ABC):
    """Abstract of token reader"""
    def __init__(self, value: Any = None):
        self.value = value

    @staticmethod
    @abstractmethod
    def read(stream: str) -> Tuple[int, Union[None, Token]]:
        pass

    def set_position(self, line, pos):
        self.line = line
        self.pos = pos

    def __eq__(self, other: Token):
        return self.__class__ == other.__class__ and self.value == other.value

    @property
    def name(self):
        return self.__class__.__name__


class TokenError(Exception):
    def __init__(self, msg: str):
        self.msg = msg

    def __str__(self):
        return f"{self.msg}"


class Number(Token):
    @staticmethod
    def read(stream: str) -> Tuple[int, Union[None, Number]]:
        string = stream[:stream.find(' #')].strip()  # ignore comment
        value = None
        for t in (int, float):
            try:
                value = t(string)
                break
            except:
                pass
        if value is not None:
            return len(string), Number(value)
        else:
            return 0, None

    @staticmethod
    def _read_number_string(stream: str):
        string_number = ""
        for c in stream[:].strip():
            if c.isnumeric():
                string_number += c
                continue
            if c == '-' and (len(string_number) == 0 or string_number[-1] == 'e'):
                string_number += c
                continue
            if c in ['e', '.'] and c not in string_number:
                string_number += c
                continue
            break
        return string_number


class Bool(Token):
    @staticmethod
    def read(stream: str) -> Tuple[int, Union[None, Bool]]:
        string = stream[:stream.find(' #')].strip()  # ignore comment
        if string in ['true', 'True', 'yes']:
            return len(string), Bool(True)
        if string in ["false", 'False', 'no']:
            return len(string), Bool(False)
        return 0, None


# class NoneType(Token):
#     @staticmethod
#     def read(stream: str) -> Tuple[int, Any]:
#         string = stream[:strip]

class String(Token):
    @staticmethod
    def read(stream: str) -> Tuple[int, Union[None, String]]:
        for op in Operator.operators:
            if stream.startswith(op):
                return 0, None
        pos = 0
        if stream[pos] in "\'\"":
            start_char = stream[pos]
            pos += 1
            while stream[pos] and stream[pos] != start_char:
                pos += 1
            if stream[pos] != start_char:
                raise TokenError("Not closed string expression")
            pos += 1
            return pos, String(stream[1:pos - 1])

        pos = len(stream)
        for op in ['#']:  # mb: other operators stops string. (hard to raise good exception)
            found_pos = stream.find(op)
            if found_pos >= 0:
                pos = min(pos, found_pos)

        if pos == 0:
            return 0, None
        return pos, String(stream[:pos].strip())


class Name(String):
    @staticmethod
    def read(stream: str) -> Tuple[int, Union[None, Name]]:
        pos = 0
        if not stream[pos].isalpha():
            return 0, None

        while pos < len(stream) and (
                stream[pos].isalnum() or stream[pos] == '_' or stream[pos] == '.'):
            pos += 1

        return pos, Name(stream[:pos])


class Operator(Token):
    operators = ['---', '@', '#', '&', '!&', '!!', '!', '- ', ': ', '*', '{', '}', '[', ']', '(', ')']

    @staticmethod
    def read(stream: str) -> Tuple[int, Union[None, Operator]]:
        for op in Operator.operators:
            if stream.startswith(op):
                return len(op), Operator(op)
        return 0, None


class Indent(Token):
    @staticmethod
    def read(stream: str) -> Tuple[int, Union[None, Indent]]:
        indent = 0
        pos = 0
        while stream[pos].isspace():
            if stream[pos] == ' ':
                indent += 1
            elif stream[pos] == '\t':
                indent += 4
            else:
                raise TokenError("Unknown indent symbol")
            pos += 1

        return pos, Indent(indent)


class List(Token):
    @staticmethod
    def read(stream: str) -> Tuple[int, Union[None, List]]:
        pos = 0
        if stream[pos] != '[':
            return pos, None
        while stream and stream[pos] != ']':
            pos += 1
        if stream[pos] != ']':
            raise TokenError('List was not closed')
        pos += 1
        read_list = eval(stream[:pos])
        return pos, List(read_list)


class TupleToken(Token):
    @staticmethod
    def read(stream: str) -> Tuple[int, Union[None, List]]:
        pos = 0
        if stream[pos] != '(':
            return pos, None
        while stream and stream[pos] != ')':
            pos += 1
        if stream[pos] != ')':
            raise TokenError('Tuple was not closed')
        pos += 1
        read_list = eval(stream[:pos])
        return pos, List(read_list)


class Dict(Token):
    @staticmethod
    def read(stream: str) -> Tuple[int, Union[None, Dict]]:
        pos = 0
        if stream[pos] != '{':
            return pos, None
        while stream and stream[pos] != '}':
            pos += 1
        if stream[pos] != '}':
            raise TokenError('Dict was not closed')
        pos += 1
        read_dict = eval(stream[:pos])
        return pos, Dict(read_dict)


class InlinePython(Token):
    @staticmethod
    def read(stream: str) -> Tuple[int, Union[None, InlinePython]]:
        pos = 0
        if stream[pos] != '`':
            return pos, None
        pos += 1
        while stream and stream[pos] != '`':
            pos += 1
        if stream[pos] != '`':
            raise TokenError('Inline python string was not clothed')
        pos += 1
        return pos, InlinePython(stream[1:pos - 1])


class PythonString(Token):
    @classmethod
    def read(cls, stream: str, implicit_fstrings: bool = False) -> \
            Tuple[int, Union[None, PythonString]]:
        string = stream[:].strip()
        if implicit_fstrings and string[0] in "\"\'":
            if string[-1] != string[0]:
                raise TokenError("Not closed f-string")
            return len(string), PythonString((string, 'f'))
        if string[0] == 'f' and string[1] in "\"\'":
            if string[-1] != string[1]:
                raise TokenError("Not closed f-string")
            return len(string), PythonString((string[1:], 'f'))
        if string[0] == 'r' and string[1] in "\"\'":
            if string[-1] != string[1]:
                raise TokenError("Not closed r-string")
            return len(string), PythonString((string[1:], 'r'))
        return 0, None



"""Contains nip directives."""

import nip.elements

from .parser import Parser, ParserError
from .constructor import Constructor
from .stream import Stream


def insert_directive(right_value, stream: Stream):
    if isinstance(right_value, nip.elements.Value):
        constructor = Constructor()
        path = constructor.construct(right_value)
        assert isinstance(path, str), "Load directive expects path as an argument."
        parser = Parser()
        config = parser.parse(path)  # Document
        return config.value

    elif isinstance(right_value, nip.elements.Args):
        assert len(right_value.value[0]) == 1, "only single positional argument will be treated as config path."
        constructor = Constructor()
        path = constructor.construct(right_value.value[0][0])
        assert isinstance(path, str), "Load directive expects path as first argument."
        parser = Parser()
        parser.link_replacements = right_value.value[1]
        config = parser.parse(path)  # Document
        return config.value

    else:
        raise ParserError(
            stream, "string or combination of arg and **kwargs are expected as value of !!insert directive")


_directives = {
    'insert': insert_directive
}


def call_directive(name, right_value, stream: Stream):
    if name not in _directives:
        raise ParserError(stream, f"Unknown parser directive '{name}'.")
    return _directives[name](right_value, stream)



from typing import List, Callable

from nip.elements import Nothing

# mb: depending on config
# mb: make Nothing a singleton
# NOTHING = None
# IS_NOTHING = lambda obj, None: obj is None

NOTHING = Nothing()
IS_NOTHING = lambda obj: isinstance(obj, Nothing)


def nothing_comparison(first, second):
    return IS_NOTHING(first) and IS_NOTHING(second)


def deep_conditioned_compare(first: object, second: object, conditions: List[Callable] = ()):
    if first.__class__ != second.__class__:  # mb: let them be subclasses of one super class
        return False
    if isinstance(first, (list, tuple)) and isinstance(second, (list, tuple)):
        if len(first) != len(second):
            return False
        return all([deep_conditioned_compare(first_item, second_item, conditions)
                    for first_item, second_item
                    in zip(first, second)])
    if isinstance(first, dict) and isinstance(second, dict):
        if len(first) != len(second):
            return False
        for key in first:
            if key not in second:
                return False
            if not deep_conditioned_compare(first[key], second[key], conditions):
                return False
    for cond in conditions:
        if cond(first, second):
            return True
    return first == second



import ast
import inspect
import typeguard

from typing import List, Dict, Union, Any


def get_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in get_subclasses(c)])


def get_sub_dict(base_cls):
    return {cls.__name__: cls for cls in get_subclasses(base_cls)}


def deep_equals(first: Union[Dict, List, Any], second: Union[Dict, List, Any]):
    if type(first) != type(second):
        return False
    if isinstance(first, dict):
        for key in first:
            if key not in second:
                return False
            if not deep_equals(first[key], second[key]):
                return False
    elif isinstance(first, list):
        if len(first) != len(second):
            return False
        for i, j in zip(first, second):
            if not deep_equals(i, j):
                return False
    else:
        return first == second
    return True


def iterate_items(obj):
    if isinstance(obj, (list, tuple)):
        for i, value in enumerate(obj):
            yield i, value
    if isinstance(obj, dict):
        for key, value in obj.items():
            yield key, value


def flatten(obj, delimiter='.', keys=()):
    if not isinstance(obj, (list, tuple, dict)):
        return {delimiter.join(keys): obj}

    result = {}
    for key, value in iterate_items(obj):
        result.update(
            flatten(value, delimiter, keys=keys + (str(key),)))
    return result




# import time
# from multiprocessing import Pool
#
#
# def f(n):
#     # time.sleep(n)
#     return n
#
#
# pool = Pool(3)
# result = pool.imap(f, (3, 1, 2))
# # result.
# print([res for res in result])



from .constructor import nip, wrap_module
from .convertor import pin
from .main import parse, parse_string, construct, load, load_string, dump, dump_string, convert, run
from .parser import Parser
from .constructor import Constructor
from .non_seq_constructor import NonSequentialConstructor



"""Contains all the elements of nip config files"""
from __future__ import annotations

import logging

import nip.constructor  # This import pattern because of cycle imports
import nip.directives
import nip.dumper
import nip.non_seq_constructor as nsc
import nip.parser
import nip.stream
import nip.tokens as tokens
import nip.utils

from abc import abstractmethod, ABC
from typing import Any, Union, Tuple, Dict

_LOGGER = logging.getLogger(__name__)


class Element(ABC):
    """Base token for nip file"""
    def __init__(self, name: str = '', value: Any = None):
        self.name = name
        self.value = value

    @classmethod
    @abstractmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Element, None]:
        pass

    def __str__(self):
        return f"{self.__class__.__name__}('{self.name}', {self.value})"

    def __getitem__(self, item):
        return self.value[item]

    def __setitem__(self, key, value):
        self.value[key] = value

    def to_python(self):
        return self.value.to_python()

    def construct(self, constructor: nip.constructor.Constructor):
        return self.value.construct(constructor)

    def dump(self, dumper: nip.dumper.Dumper):
        return self.value.dump(dumper)

    def __eq__(self, other):
        return self.name == other.name and self.value == other.value

    def flatten(self, delimiter='.') -> Dict:
        return nip.utils.flatten(self.to_python(), delimiter)


class Document(Element):  # ToDo: add multi document support
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Document:
        doc_name = cls._read_name(stream)
        content = RightValue.read(stream, parser)
        return Document(doc_name, content)

    @classmethod
    def _read_name(cls, stream: nip.stream.Stream):
        read_tokens = stream.peek(tokens.Operator('---'), tokens.Name) or \
                      stream.peek(tokens.Operator('---'))
        if read_tokens is not None:
            stream.step()
            if len(read_tokens) == 2:
                return read_tokens[1].value
        return ''

    def dump(self, dumper: nip.dumper.Dumper):
        string = "---"
        if self.name:
            string += " " + self.name + " "
        return string + self.value.dump(dumper)


class RightValue(Element):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Element:
        value = Directive.read(stream, parser) or \
                LinkCreation.read(stream, parser) or \
                Link.read(stream, parser) or \
                Class.read(stream, parser) or \
                Tag.read(stream, parser) or \
                Iter.read(stream, parser) or \
                Args.read(stream, parser) or \
                FString.read(stream, parser) or \
                Nothing.read(stream, parser) or \
                InlinePython.read(stream, parser) or \
                Value.read(stream, parser)

        if value is None:
            raise nip.parser.ParserError(stream, "Wrong right value")

        return value


class Value(Element):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[None, Value]:
        tokens_list = [
            tokens.Number,
            tokens.Bool,
            tokens.String,
            tokens.List,
            tokens.TupleToken,
            tokens.Dict
        ]
        for token in tokens_list:
            read_tokens = stream.peek(token)
            if read_tokens is not None:
                stream.step()
                return Value(read_tokens[0].name, read_tokens[0].value)

        return None

    def to_python(self):
        return self.value

    def construct(self, constructor: nip.constructor.Constructor):
        return self.value

    def dump(self, dumper: nip.dumper.Dumper):
        if isinstance(self.value, str):
            return f'"{self.value}"'
        return str(self.value)

    def __len__(self):
        return len(self.value)


class LinkCreation(Element):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Element, None]:
        read_tokens = stream.peek(tokens.Operator('&'), tokens.Name)
        if read_tokens is None:
            # if stream.peek(tokens.Operator('&')):  # mb: do it more certainly: peak operator
            #     raise nip.parser.ParserError(      # mb: firstly and then choose class to read)
            #         stream, "Found variable creation operator '&' but name is not specified")
            return None

        name = read_tokens[1].value
        stream.step()

        value = RightValue.read(stream, parser)
        if name in parser.links:
            raise nip.parser.ParserError(stream, f"Redefining of link '{name}'")
        parser.links.append(name)

        return LinkCreation(name, value)

    def construct(self, constructor: nip.constructor.Constructor):
        if nsc.should_construct(self.name, constructor):
            constructor.vars[self.name] = self.value.construct(constructor)
        return constructor.vars[self.name]

    def dump(self, dumper: nip.dumper.Dumper):
        return f"&{self.name} {self.value.dump(dumper)}"


class Link(Element):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Element, None]:
        read_tokens = stream.peek(tokens.Operator('*'), tokens.Name)
        if read_tokens is None:
            return None

        name = read_tokens[1].value
        stream.step()

        if name in parser.link_replacements:
            return parser.link_replacements[name]

        if parser.sequential_links and name not in parser.links:
            nip.parser.ParserError(stream, "Link usage before assignment")

        return Link(name)

    def to_python(self):
        return "nil"  # something that means that object is not constructed yet.

    def construct(self, constructor: nip.constructor.Constructor):
        return constructor.vars[self.name]

    def dump(self, dumper: nip.dumper.Dumper):
        return f"*{self.name}"


class Tag(Element):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Tag, None]:
        read_tokens = stream.peek(tokens.Operator('!'), tokens.Name)
        if read_tokens is None:
            return None
        name = read_tokens[1].value
        stream.step()

        value = RightValue.read(stream, parser)

        return Tag(name, value)

    def construct(self, constructor: nip.constructor.Constructor):
        if isinstance(self.value, Args):
            args, kwargs = self.value.construct(constructor, always_pair=True)
        else:
            value = self.value.construct(constructor)
            if isinstance(value,  Nothing):  # mb: Add IS_NOTHING method
                return constructor.builders[self.name]()
            else:
                args, kwargs = [value], {}

        if self.name not in constructor.builders:
            raise nip.constructor.ConstructorError(
                self, args, kwargs, f"Constructor for Tag '{self.name}' is not registered.")

        messages = nip.constructor.check_typing(constructor.builders[self.name], args, kwargs)
        if len(messages) > 0:
            if constructor.strict_typing:
                raise nip.constructor.ConstructorError(self, args, kwargs, "\n".join(messages))
            else:
                _LOGGER.warning(f"Typing mismatch while constructing {self.name}:\n" +
                                "\n".join(messages))

        try:  # Try to construct
            return constructor.builders[self.name](*args, **kwargs)
        except Exception as e:
            raise nip.constructor.ConstructorError(self, args, kwargs, e)

    def dump(self, dumper: nip.dumper.Dumper):
        return f"!{self.name} " + self.value.dump(dumper)


class Class(Element):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Class, None]:
        read_tokens = stream.peek(tokens.Operator('!&'), tokens.Name)
        if read_tokens is None:
            return None
        name = read_tokens[1].value
        stream.step()

        value = RightValue.read(stream, parser)
        if not isinstance(value, Nothing):
            raise nip.parser.ParserError(stream, "Class should be created with nothing to the right.")

        return Class(name, value)

    def construct(self, constructor: nip.constructor.Constructor):
        value = self.value.construct(constructor)
        assert isinstance(value, Nothing), "Unexpected right value while constructing Class"

        return constructor.builders[self.name]

    def dump(self, dumper: nip.dumper.Dumper):
        return f"!&{self.name} " + self.value.dump(dumper)


class Args(Element):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Args, None]:
        start_indent = stream.pos
        if start_indent <= parser.last_indent:
            return None

        args = []
        kwargs = {}  # mb: just dict with integer and string keys
        read_kwarg = False
        while stream and stream.pos == start_indent:
            parser.last_indent = start_indent

            item = cls._read_list_item(stream, parser)
            if item is not None:
                if parser.strict and read_kwarg:
                    raise nip.parser.ParserError(stream,
                                                 "Positional argument after keyword argument "
                                                 "is forbiddent in `strict` mode.")
                args.append(item)
                continue

            key, value = cls._read_dict_pair(stream, parser, kwargs.keys())
            if key is not None:
                read_kwarg = True
                kwargs[key] = value
                continue

            break

        if stream.pos > start_indent:
            raise nip.parser.ParserError(stream, "Unexpected indent")

        if not args and not kwargs:
            return None
        return Args("args", (args, kwargs))

    @classmethod
    def _read_list_item(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) \
            -> Union[Element, None]:
        read_tokens = stream.peek(tokens.Operator('- '))
        if read_tokens is None:
            return None
        stream.step()

        value = RightValue.read(stream, parser)

        return value

    @classmethod
    def _read_dict_pair(cls, stream: nip.stream.Stream, parser: nip.parser.Parser, kwargs_keys) \
            -> Union[Tuple[str, Element], Tuple[None, None]]:
        # mb: read String instead of Name for keys with spaces,
        # mb: but this leads to the case that
        read_tokens = stream.peek(tokens.Name, tokens.Operator(': '))
        if read_tokens is None:
            return None, None

        key = read_tokens[0].value
        if parser.strict and key in kwargs_keys:
            raise nip.parser.ParserError(stream,
                                         f"Dict key overwriting is forbidden in `strict` "
                                         f"mode. Overwritten key: '{key}'.")
        stream.step()

        value = RightValue.read(stream, parser)

        return key, value

    def __str__(self):
        args_repr = "[" + ", ".join([str(item) for item in self.value[0]]) + "]"
        kwargs_repr = \
            "{" + ", ".join([f"{key}: {str(value)}" for key, value in self.value[1].items()]) + "}"

        return f"{self.__class__.__name__}('{self.name}', {args_repr}, {kwargs_repr})"

    def __bool__(self):
        return bool(self.value[0]) or bool(self.value[1])

    def __getitem__(self, item):
        try:
            return self.value[0][item]
        except TypeError:
            return self.value[1][item]

    def __setitem__(self, key, value):
        if isinstance(key, int):
            self.value[0][key] = nip.convert(value)
        else:
            self.value[1][key] = nip.convert(value)

    def append(self, value):
        self.value[0].append(nip.convert(value))

    def __len__(self):
        return len(self.value[0]) + len(self.value[1])

    def __iter__(self):
        for item in self.value[0]:
            yield item
        for key, item in self.value[1].items():
            yield item

    def to_python(self):
        args = list(item.to_python() for item in self.value[0])
        kwargs = {key: value.to_python() for key, value in self.value[1].items()}
        assert args or kwargs, "Error converting Args node to python"  # This should never happen
        if args and kwargs:
            result = {}
            result.update(nip.utils.iterate_items(args))
            result.update(nip.utils.iterate_items(kwargs))
            return result
        return args or kwargs

    def construct(self, constructor: nip.constructor.Constructor, always_pair=False):
        args = list(item.construct(constructor) for item in self.value[0])
        kwargs = {key: value.construct(constructor) for key, value in self.value[1].items()}
        assert args or kwargs, "Error converting Args node to python"  # This should never happen
        if args and kwargs or always_pair:
            return args, kwargs
        return args or kwargs

    def dump(self, dumper: nip.dumper.Dumper):
        dumped_args = '\n'.join([
            " "*dumper.indent + f"- {item.dump(dumper + dumper.default_shift)}"
            for item in self.value[0]
        ])
        string = ('\n' if dumped_args else '') + dumped_args

        dumped_kwargs = '\n'.join([
            " "*dumper.indent + f"{key}: {value.dump(dumper + dumper.default_shift)}"
            for key, value in self.value[1].items()
        ])
        string += ('\n' if dumped_kwargs else '') + dumped_kwargs

        return string


class Iter(Element):
    def __init__(self, name: str = '', value: Any = None):
        super(Iter, self).__init__(name, value)
        self.return_index = -1
        # mb: name all the iterators and get the value from constructor rather then use this index

    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Iter, None]:
        read_tokens = stream.peek(tokens.Operator('@'), tokens.Name) or \
                      stream.peek(tokens.Operator('@'))
        if read_tokens is None:
            return None
        stream.step()
        value = RightValue.read(stream, parser)
        if isinstance(value, Value) and isinstance(value.value, list):
            value = value.value
        elif isinstance(value, Args) and len(value.value[1]) == 0:
            value = value
        else:
            raise nip.parser.ParserError(stream, "List is expected as a value for Iterable node")
        if len(read_tokens) == 1:
            iterator = Iter('', value)
        else:
            iterator = Iter(read_tokens[1].value, value)

        parser.iterators.append(iterator)
        return iterator

    def to_python(self):
        if self.return_index == -1:
            raise iter(self.value)
        if isinstance(self.value[self.return_index], Element):
            return self.value[self.return_index].to_python()
        return self.value[self.return_index]

    def construct(self, constructor: nip.constructor.Constructor):
        if self.return_index == -1:
            raise Exception("Iterator index was not specified by IterParser")
        if isinstance(self.value, list):
            return self.value[self.return_index]
        elif isinstance(self.value, Args):
            return self.value[self.return_index].construct(constructor)
        else:
            raise nip.constructor.ConstructorError(self, (), {}, "Unexpected iter value type")

    def dump(self, dumper: nip.dumper.Dumper):
        if self.return_index == -1:
            raise nip.dumper.DumpError(
                "Dumping an iterator but index was not specified by IterParser"
            )
        if isinstance(self.value, list):
            return str(self.value[self.return_index])
        elif isinstance(self.value, Args):
            return self.value[self.return_index].dump(dumper)
        else:
            raise nip.dumper.DumpError("Unable to dump Iterable node: unexpected value type")


class InlinePython(Element):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> \
            Union[InlinePython, None]:
        read_tokens = stream.peek(tokens.InlinePython)
        if read_tokens is None:
            return None
        stream.step()
        exec_string = read_tokens[0].value
        return InlinePython(value=exec_string)

    def construct(self, constructor: nip.constructor.Constructor):
        nsc.preload_vars(self.value, constructor)
        locals().update(constructor.vars)
        return eval(self.value)

    def dump(self, dumper: nip.dumper.Dumper):
        return f"`{self.value}`"

    def to_python(self):
        return f"`{self.value}`"


class Nothing(Element):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Nothing, None]:
        if not stream:
            return Nothing()

        indent = stream.pos
        if stream.pos == 0 or (
                stream.lines[stream.n][:stream.pos].isspace()
                and indent <= parser.last_indent):
            return Nothing()

    def construct(self, constructor: nip.constructor.Constructor):
        return self

    def dump(self, dumper: nip.dumper.Dumper):
        return ""

    def to_python(self):
        return None


class FString(Element):  # Includes f-string and r-string
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> \
            Union[FString, None]:
        read_tokens = stream.peek(tokens.PythonString)
        if read_tokens is None:
            return None
        stream.step()
        string, t = read_tokens[0].value
        if t == 'r':
            print("Warning: all strings in NIP are already python r-string. "
                  "You don't have to explicitly specify it.")
        return FString(value=string)

    def construct(self, constructor: nip.constructor.Constructor):
        nsc.preload_vars(f"f{self.value}", constructor)
        locals().update(constructor.vars)
        return eval(f"f{self.value}")

    def dump(self, dumper: nip.dumper.Dumper):
        return f"f{self.value}"

    def to_python(self):
        return f"f{self.value}"


class Directive(Element):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> \
            Union[FString, None]:
        read_tokens = stream.peek(tokens.Operator('!!'), tokens.Name)
        if read_tokens is None:
            return None
        name = read_tokens[1].value
        stream.step()

        value = RightValue.read(stream, parser)

        return nip.directives.call_directive(name, value, stream)



import symtable

import nip.elements

from .constructor import Constructor, ConstructorError


class VarsDict:
    def __init__(self, constructor):
        super().__init__()
        self.constructor = constructor
        self.vars = {}
        self.in_progress = set()

    def __getitem__(self, item):
        if item in self.vars:
            return self.vars[item]
        if item not in self.constructor.links:
            raise NonSequentialConstructorError(f"Unresolved link '{item}'")
        if item in self.in_progress:
            raise NonSequentialConstructorError(f"Recursive construction of '{item}'.")
        self.in_progress.add(item)
        self.constructor.links[item].construct(self.constructor)
        self.in_progress.remove(item)
        return self.vars[item]

    def __setitem__(self, key, value):
        if key not in self.vars:  # was not constructed earlier
            self.vars[key] = value

    def __iter__(self):
        return iter(self.vars)

    def items(self):
        return self.vars.items()

    def keys(self):
        return self.vars.keys()


class NonSequentialConstructor(Constructor):
    def __init__(self, base_config: "nip.elements.Element",
                 ignore_rewriting=False,
                 load_builders=True,
                 strict_typing=False):
        super().__init__(ignore_rewriting, load_builders, strict_typing)
        self.vars = VarsDict(self)
        self.links = {}
        self._find_links(base_config)

    def _find_links(self, node: "nip.elements.Element"):
        if isinstance(node, nip.elements.LinkCreation):
            assert node.name not in self.links, "Redefined link."
            self.links[node.name] = node
        if isinstance(node, nip.elements.Args):
            for sub_node in node:
                self._find_links(sub_node)
        if isinstance(node.value, nip.elements.Element):
            self._find_links(node.value)


class NonSequentialConstructorError(Exception):
    def __init__(self, massage):
        self.massage = massage

    def __str__(self):
        return self.massage


def preload_vars(code, constructor: Constructor):
    if not isinstance(constructor, NonSequentialConstructor):
        return
    table = symtable.symtable(code, 'string', 'exec')
    for name in constructor.links:
        try:
            if table.lookup(name).is_global():
                constructor.vars[name]
        except KeyError:
            pass


def should_construct(name, constructor: Constructor):
    if isinstance(constructor, NonSequentialConstructor):
        return not (name in constructor.vars)
    return True  # we can reconstruct



# Constructor of tagged objects
import importlib
import importlib.util
import inspect
import typeguard

from typing import Callable, Optional, Union, List, Tuple
from types import FunctionType, ModuleType, BuiltinFunctionType
from multipledispatch import dispatch

from .utils import get_sub_dict


global_builders = {}  # builders shared between Constructors
global_calls = {}  # history of object creations


class Constructor:
    def __init__(self, ignore_rewriting=False, load_builders=True,
                 strict_typing=False):
        self.builders = {}
        self.ignore_rewriting = ignore_rewriting
        if load_builders:
            self.load_builders()
        self.vars = {}
        self.strict_typing = strict_typing

    def construct(self, element):
        return element.construct(self)

    def register(self, func: Callable, tag: Optional[str] = None):
        """Registers builder function for tag
        Parameters
        ----------
        func:
            Function or class to build the python object.
            In case of class its __init__ method will be called to construct object.
        tag: str, optional
            Tag in yaml/nip file. func.__name__ will be used if not specified.
        """
        if tag is None:
            tag = func.__name__
        assert self.ignore_rewriting or tag not in self.builders, \
            f"Builder for tag '{tag}' already registered"
        self.builders[tag] = func

    def load_builders(self):
        self.builders.update(global_builders)
        self.builders.update(get_sub_dict(NIPBuilder))


class ConstructorError(Exception):
    def __init__(self, element, args, kwargs, e):
        self.cls = type(element).__name__
        self.name = element.name
        self.args = args
        self.kwargs = kwargs
        self.e = e

    def __str__(self):
        return f"Unable to construct {self.cls} '{self.name}' with args:{self.args} and " \
               f"kwargs:{self.kwargs}.\nFollowing exception occurred:\n" \
               f"{self.e.__class__.__name__}: {self.e}"


# mb: add meta for auto detecting this class as NIP-builder
# ToDo: Add init wrapper for auto detection init args for convenient object dumping
class NIPBuilder:
    pass


def nip_decorator(name=None, wrap_call=False):
    assert name is None or len(name) > 0, "name should be nonempty"

    def _(item):
        if wrap_call:
            assert isinstance(item, type), "Call wrapping supported only for class type"
            item = call_wrapper(item)
        local_name = name or item.__name__
        if isinstance(local_name, (list, tuple)):
            for n in local_name:
                global_builders[n] = item
        else:
            global_builders[local_name] = item
        return item

    return _


@dispatch(str)
def nip(name: str, wrap_call: bool = False):
    """Wrapper to register function or class with specified name."""
    return nip_decorator(name, wrap_call)


@dispatch([(list, tuple)])
def nip(names: Union[List[str], Tuple[str]], wrap_call: bool = False):
    """Wrapper to register function or class with specified name."""
    assert len(names) > 0, "Tag list should be nonempty"
    for name in names:
        assert isinstance(name, str), "Every specified Tag should be string."
    return nip_decorator(names, wrap_call)


@dispatch()
def nip(wrap_call=False):
    return nip_decorator(wrap_call=wrap_call)


@dispatch([(type, FunctionType, BuiltinFunctionType)])
def nip(item: Union[type, FunctionType]):
    """Wrapper to register function or class with default name."""
    return nip_decorator()(item)


@dispatch(ModuleType)
def nip(module, wrap_builtins=False):
    """Wraps everything specified under module."""
    wrap_module(module, wrap_builtins)


def call_wrapper(item: Union[type, FunctionType]):  # wraps call for convenient object dump
    print("wrapping")
    # ToDo: only for classes !!

    def call_dumper(*args, **kwargs):
        print(args, kwargs)
        value = item(*args, **kwargs)
        global_calls[value] = (args, kwargs)  # Todo: how to do this correctly?
        return value

    call_dumper.__doc__ = item.__doc__  # mimic to original function
    call_dumper.__name__ = item.__name__

    return call_dumper


def wrap_module(module: Union[str, ModuleType], wrap_builtins=False):
    """Wraps everything declared in module with @nip

    Parameters
    ----------
    module: str or ModuleType
        Module name (e.g. "numpy.random") or module itself
    wrap_builtins
        Whether to wrap builtin functions or not.
        (Useful when wrapping whole module like `numpy`)
    """
    if isinstance(module, str):
        module = importlib.import_module(module)

    for value in module.__dict__.values():
        if isinstance(value, (type, FunctionType)) or \
                wrap_builtins and isinstance(value, BuiltinFunctionType):
            nip(value)


def check_typing(func, args, kwargs) -> List[str]:
    try:
        signature = inspect.signature(func)
    except ValueError:  # unable to load function signature (common case for builtin functions)
        return []
    messages = []
    for arg, param in zip(args, signature.parameters.values()):
        if param.annotation is inspect.Parameter.empty:
            continue
        try:
            typeguard.check_type(arg, param.annotation)
        except typeguard.TypeCheckError as e:
            messages.append(f"{param.name}: {e}")

    for name, value in kwargs.items():
        if name not in signature.parameters:
            continue  # handled by python
        annotation = signature.parameters[name].annotation
        if annotation is inspect.Parameter.empty:
            continue
        try:
            typeguard.check_type(value, annotation)
        except typeguard.TypeCheckError as e:
            messages.append(f"{name}: {e}")

    return messages
